import sys
import os
import subprocess
import argparse
from . import file as jsf
from . import run_all
from . import generate_form
from . import functions as func
from timeit import default_timer as timer


def Model_is_EFT(model):
    # checks if model/QGRAF/model_data/listheavy is empty
    # if it is the model is an EFT (no heavy particles)
    # if it isn't the model is a UV model (with heavy particles)
    # it returns True or False
    with open(os.path.join(model,"QGRAF","model_data","listheavy")) as fi:
        s = fi.read()
    return(len(s)==0)


def check_linear_depencence_amplitudes(EFTmodel,wolframcommand):
    EFTmodeldir=jsf.proper_dir_path(EFTmodel)
    # first we check that we are using an EFT
    if Model_is_EFT(EFTmodel):
        # Amplitudes needed to match onto the EFT
        inputfile=os.path.join(EFTmodel,"QGRAF","model_data","allamplitudes.txt")
        # Check if inputfile exists otherwise create it
        if not os.path.isfile(inputfile):
            func.generate_allamplitudes(EFTmodel)
        #check we are dealing with an eft model
        if Model_is_EFT(EFTmodel):
            start=timer()
            print("Computing amplitudes to check linear independence of operators in "+EFTmodel+". This might take some time depending on the complexity of the model.")
            # we only do EFT at tree level with very large maxdim
            try:
                run_all.run_all(inputfile,EFTmodeldir,True,20)
            except Exception as E:
                raise Exception(E)

            if wolframcommand!=[]:
                coredir=func.get_path("core")
                createscript=os.path.join(coredir,"check_linear_dependence.wl")
                print("Checking linear independence of the operators in EFT model "+EFTmodel+". This might take some time depending on the complexity of the model") 
                start = timer()
                subprocess.run(wolframcommand+[createscript, EFTmodel])
                end = timer()
                print("Check finished. It took "+str(int(round(end-start)))+" seconds to compute them.")


        else:
            print("The model needs to be an EFT")

    else:
        print("The argument of this functions has to be an EFT")    
        
def match_model(NPmodel,EFTmodel,onlytree=False):
    NPmodeldir=jsf.proper_dir_path(NPmodel)
    EFTmodeldir=jsf.proper_dir_path(EFTmodel)
    # first we check NPmodel has heavy particles
    # and EFTmodel has no heavy particles
    if Model_is_EFT(EFTmodel):
        # Amplitudes needed to match onto the EFT
        inputfile=os.path.join(EFTmodel,"QGRAF","model_data","amplitudes.txt")
        # Check if inputfile exists otherwise create it
        if not os.path.isfile(inputfile):
            func.generate_amplitudes(EFTmodel)
        if Model_is_EFT(NPmodel):
            # UV model is also an EFT
            # we allow now a different EFT for generation and matching
            # so that we can do just partial EFT integration
            if True: #NPmodel==EFTmodel:
                # RGE
                print("Running in RGEMaker mode")

                start=timer()
                print("Computing amplitudes to match the UV poles of "+EFTmodel+" onto "+NPmodel+". This might take some time depending on the complexity of the model.")
                # we start with the EFT at tree level and maxdim very large
                try:
                    run_all.run_all(inputfile,EFTmodeldir,True,20)
                except Exception as E:
                    raise Exception(E)
                maxdim=generate_form.get_maxdim(EFTmodeldir)
                # we now do the do again the EFT, with maxdim fixed both tree level and one loop
                # we could do it directly at one loop but it does not repeat the tree level amplitudes
                # already computed
                # note that isEFT is set to false but isnoteft internally will be also set to false (0)
                if onlytree:
                    try:
                        run_all.run_all(inputfile,NPmodeldir,True,maxdim)
                    except Exception as E:
                        raise Exception(E)

                else:
                    try:
                        run_all.run_all(inputfile,NPmodeldir,False,maxdim)
                    except Exception as E:
                        raise Exception(E)
                end = timer()
                print("Amplitudes to match model "+NPmodel+" onto EFT "+EFTmodel+" computed.")
                print ("time taken "+str(int(round(end-start)))+" seconds ")
                
                
            else:
                print("We cannot match an EFT to a different EFT")
                return()
        else:
            # Matching a UV model to an EFT
            start=timer()
            print("Computing amplitudes to match model "+NPmodel+" onto EFT "+EFTmodel+". This might take some time depending on the complexity of the model.")
            # we start with the EFT, we only do tree level and maxdim very large
            try:
                run_all.run_all(inputfile,EFTmodeldir,True,20)
            except Exception as E:
                raise Exception(E)
            maxdim=generate_form.get_maxdim(EFTmodeldir)
            # we now do the UV model, if onlytree=True we do only tree level
            if onlytree:
                try:
                    run_all.run_all(inputfile,NPmodeldir,True,maxdim)
                except Exception as E:
                    raise Exception(E)
            # we now do the UV model, if onlytree=False we do both tree level and one loop
            else:
                try:
                    run_all.run_all(inputfile,NPmodeldir,False,maxdim)
                except Exception as E:
                    raise Exception(E)
            end = timer()
            print("Amplitudes to match model "+NPmodel+" onto EFT "+EFTmodel+" computed.")
            print ("time taken "+str(int(round(end-start)))+" seconds ")
            
    else:
        print("The second model has to be an EFT")
        return()
    



import sys
import os
import subprocess

from . import file as jsf
from . import run_qgraf 
from . import generate_form 
from . import run_form 
from . import functions as func



def run_all(inputfile,modeldir0,isEFT,maxdim):

    modeldir=jsf.proper_dir_path(modeldir0)
    if os.path.isfile(inputfile):
        if os.path.isdir(modeldir):
            # Run QGRAF
            try:
                run_qgraf.run_qgraf(inputfile,modeldir,isEFT)
            except Exception as E:
                raise Exception(E)
            # Generate Form
            generate_form.generate_form(modeldir,isEFT,maxdim)
            # Run Form
            try:
                run_form.run_form(modeldir,isEFT)
            except Exception as E:
                raise Exception(E)
        else:
            print("Model directory "+modeldir+" does not exist")
    else:
        print("Input file "+inputfile+" does not exist")



import os
import cmd
import sys
import readline
import glob
import pickle
import configparser
import inspect
import pkg_resources  
import shlex
from pathlib import Path
from . import functions
from . import match_model
from . import init_mm




__init_data = init_mm.initialise_mm()


def create_model(args=""):
    """Generates the matchmaker-eft model from a FeynRules one"""
    if len(args) == 0:
        print("We need the address of the new physics model that you want to match")
    else:
        try:
            functions.create_model(args.split(),__init_data["wc"],__init_data["fr_path"])
        except Exception as E:
            raise(E)

def clean_model(args=""):
    """Removes all calculations from previously matched model so that it can be matched again"""
    if len(args) == 0:
        print("We need the address of the new physics model that you want to clean")
    else:
        for fi in glob.glob(os.path.join(args,"*","proc*","*")):
            os.remove(fi)
        for fi in ['MatchingResult.dat','MatchingProblems.dat','RGEResult.dat']:
            if os.path.isfile(os.path.join(args,fi)):
                os.remove((os.path.join(args,fi)))

def download_models(args=""):
    """Downloads the file MatchMaker.tar.gz with sample FeynRules models and a Mathematica notebook to create MatchMaker models"""
    if os.path.isdir(args):
        functions.download_models(args)
    else:
        print("We need the path of the directory where you want the models to be downloaded")

def check_linear_dependence(args=""):
    """Checks linear independence of the operators defined in an EFT."""
    if len(args) == 0:
        print("We need the address of the new physics model that you want to match")
    else:
        try:
            match_model.check_linear_depencence_amplitudes(args, __init_data["wc"])
        except Exception as E:
            print(E)

def match_model_to_eft_amplitudes(args=""):
    """Provides the complete one-loop matching between any matchmaker-eft UV model onto any matchmaker-eft eft."""
    if len(args) == 0:
        print("We need the address of the new physics model that you want to match")
    else:
        try:
            match_model.match_model(*args.split())
        except Exception as E:
            # inspect.stack()[1].function is the function that calls match_model_to_eft_amplitudes
            # we raise an exception only when it is called from that function
            callingfunction=inspect.stack()[1].function
            if callingfunction in ["match_model_to_eft","compute_rge_model_to_eft"]:
                print(E)
                raise(E)
            print(E)


def match_model_to_eft_amplitudes_onlytree(args=""):
    """Provides the complete one-loop matching between any matchmaker-eft UV model onto any matchmaker-eft eft."""
    if len(args) == 0:
        print("We need the address of the new physics model that you want to match")
    else:
        try:
            match_model.match_model(*args.split(),True)
        except Exception as E:
            # inspect.stack()[1].function is the function that calls match_model_to_eft_amplitudes
            # we raise an exception only when it is called from that function
            if inspect.stack()[1].function == "match_model_to_eft_onlytree":
                print(E)
                raise(E)
            print(E)


def compute_wilson_coefficients(args=""):
    """Computes the Wilson coefficients for an already matched MatchMaker model"""
    if len(args) == 0:
        print("We need the address of the new physics model that you want to match")
    else:
        functions.compute_wilson_coefficients(*args.split(), __init_data["wc"])

def compute_rge_model_to_eft(args=""):
    """Computes the Wilson coefficients for an already matched MatchMaker model"""
    if len(args) == 0:
        print("We need the address of the new physics model that you want to match")
    else:
        try:
            match_model_to_eft_amplitudes(args)
            functions.compute_rge(*args.split(), __init_data["wc"])
        except Exception as E:
            raise(E)


def match_model_to_eft(args=""):
    """Provides the complete one-loop matching between any matchmaker-eft UV model onto any matchmaker-eft eft."""
    try:
        match_model_to_eft_amplitudes(args)
        compute_wilson_coefficients(args)
    except Exception as E:
        pass



def match_model_to_eft_onlytree(args=""):
    """Provides the complete one-loop matching between any matchmaker-eft UV model onto any matchmaker-eft eft."""
    try:
        match_model_to_eft_amplitudes_onlytree(args)
        compute_wilson_coefficients(args)
    except Exception as E:
        pass




from . import file as jsf
import os
import subprocess
import sys
import glob
from colorama import Fore
from colorama import Style
from . import functions as func


def write_one_point_function(partlist,dirname):
    # QGRAF does not do tree-level one point functions (tadpoles) so we do them ourselves
    # partlist is a list of length 1 with the name of the particle involved
    # we only do it at tree level if there is an entry in v1list
    resp="diags:=[\nNULL]:";
    if os.path.isfile(dirname+'QGRAF/model_data/v1list'):
        with open(dirname+'QGRAF/model_data/v1list','r') as the_file:
            lines = the_file.read().strip().splitlines()
        v1parts=[[li.split()[0]] for li in lines]
        if (partlist in v1parts):
            resp="diags:=[\n\n(+1)\n cpol("+partlist[0]+"(-1,p1))*\n\n v1("+partlist[0]+"(-1,p1)),\nNULL]:"
    return resp
    
def write_two_point_function(partlist,dirname):
    # QGRAF does not do tree-level two point functions so we do them ourselves
    # partlist is a list of length 2 with the names of the two particles involved
    # we only do it for the effective theory, that is for models with no heavy particles
    # as for models with heavy particles there is no tree level contribution that involves
    # heavy particles

    #get the sign
    # we go through the propagators and get the sign
    with open(dirname+'QGRAF/model','r') as the_file:
        lines = the_file.read().strip().splitlines()
    thelines=[li.strip().replace('[','').replace(']','').split(',') for li in lines if '+' in li or '-' in li]
    signo=''

    #get the sign
    # we go through the propagators and get the sign
    with open(dirname+'QGRAF/model_data/listlightfermions','r') as the_file:
        fermionlines = the_file.read().strip().replace(' ','').split(',')
    with open(dirname+'QGRAF/model_data/listlight','r') as the_file:
        lines = the_file.read().strip().replace(' ','').split(',')
    if partlist[0] in fermionlines or partlist[1] in fermionlines:
        signo='(-1)*'
    elif partlist[0] in lines or partlist[1] in lines:
        signo='(+1)*'

    with open(dirname+"QGRAF/model_data/listheavy","r") as the_file:
        # check if there are heavy particles
        # noheavy=True if there are no heavy particles in the spectrum
        lines = the_file.read().strip().splitlines()
        noheavy=len(lines)==0

    if len(signo)>1:
        # we only proceed if the sign has been computed properly
        # we now check the order of the 2-point vertex
        # read two-point vertices
        with open(dirname+'QGRAF/model_data/v2list','r') as the_file:
            lines = the_file.read().strip().splitlines()
        v2parts=[(li.split())[:2] for li in lines]

        #if (partlist in v2parts) and noheavy:
        if (partlist in v2parts):
            #correct order
            return "diags:=[\n\n"+signo+"\n cpol("+partlist[0]+"(-1,p1))*\n cpol("+partlist[1]+"(-3,p2))*\n\n v2("+partlist[0]+"(-1,p1),"+partlist[1]+"(-3,p2)),\nNULL]:"
        elif (list(reversed(partlist)) in v2parts):
            #opposite order
            if '-' in signo:
                # sign is -*-=+ if in reverse order 
                signo='(+1)*'
            return "diags:=[\n\n"+signo+"\n cpol("+partlist[0]+"(-1,p1))*\n cpol("+partlist[1]+"(-3,p2))*\n\n v2("+partlist[1]+"(-3,p2),"+partlist[0]+"(-1,p1)),\nNULL]:"
        else:
            return "diags:=[\nNull]:"
    else:
        return signo

def run_qgraf(listofparticles,dirname0,isEFT):
    if os.path.isdir(dirname0):
        if os.path.isfile(listofparticles):
            dirname=jsf.proper_dir_path(dirname0)
            if isEFT:
                filelist=glob.glob(dirname+"QGRAF/qgraf.skeleton.0loop.dat")
            else:
                filelist=glob.glob(dirname+"QGRAF/qgraf.skeleton.*loop.dat")                
            with open(listofparticles) as f:
                lines = f.read().strip().splitlines()
            for li in lines:
                outfile=" ".join(li.split()).replace(" ","_").replace("[","").replace("]","")
                partlist=','.join(li.split())
                for fil in filelist:
                    outputdirname=dirname+"QGRAF/proc_"+fil.split('/')[-1].split('.')[-2]+"/"
                    if not os.path.isdir(outputdirname):
                        subprocess.run(("mkdir "+outputdirname).split())
                    outputfilename=dirname+"QGRAF/proc_"+fil.split('/')[-1].split('.')[-2]+"/"+outfile+".qgf"
                    
                    if not os.path.isfile(outputfilename) and "0" in fil.split('.')[-2] and len(li.split()) == 2:
                        # qgraf does not do tree level two point functions, we do it ourselves
                        twop=write_two_point_function(li.split(),dirname)
                        if len(twop)>1:
                            with open(outputfilename,'w') as the_file:
                                the_file.write(twop)
                    elif not os.path.isfile(outputfilename) and "0" in fil.split('.')[-2] and len(li.split()) == 1:
                        # qgraf does not do tree level one point functions, we do it ourselves
                        onep=write_one_point_function(li.split(),dirname)
                        if len(onep)>1:
                            with open(outputfilename,'w') as the_file:
                                the_file.write(onep)
                    elif not os.path.isfile(outputfilename):
                        themodel=dirname+"QGRAF/model"
                        subprocess.run(("cp "+themodel+" ./model").split())
                        myrepl=("LISTOFPARTICLES",partlist),("THEMODEL",'model')
                        jsf.dress_skeleton(myrepl,fil,'qgraf.dat')
                        func.rep_mysty('qgraf.dat')
                        if os.path.isfile('OUTPUTFILE'):
                            subprocess.run(("rm OUTPUTFILE").split())
                        p=subprocess.run(("qgraf").split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
                        logfile=outputfilename[:-3]+'log'
                        f = open(logfile, "w") 
                        f.write(p.stdout.decode())
                        f.close()
                        if p.stdout.decode().find('error') > 0:
                            subprocess.run(("rm model").split())
                            subprocess.run(("rm qgraf.dat").split())
                            level=fil.split('/')[-1].split('.')[-2]
                            raise Exception("\n"+ Fore.RED +"There was an error with QGRAF computing the amplitude "+outputfilename.split('/')[-1][:-4].replace('_',' ')+" at "+level+". See the log file "+logfile+Style.RESET_ALL+"\n")
                        else: 
                            subprocess.run(("mv OUTPUTFILE "+outputfilename).split())
                            subprocess.run(("rm model").split())
                            subprocess.run(("rm qgraf.dat").split())
        else:
            print("not a valid list of particles")
    else:
        print("not a valid directory")





import shutil
import os
import subprocess
import datetime
import tarfile
from . import functions as func
from . import match_model


def ts_up(dest):
    #go one directory up (to the original working directory)
    os.chdir(os.path.dirname(os.getcwd()))
    # #once we are done we delete the temporary directory
    # if os.path.isdir(dest):
        # shutil.rmtree(dest)

def test_installation(wolframcommand, frpath):
    print ("Starting the tests at ",datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"))
    src_path=func.get_path("data")
    src_fullname=os.path.join(src_path,"test_suite.tar.gz")
    # dest="tempdir"+''.join(random.choice(string.ascii_lowercase) for i in range(5))
    dest=".test_suite"
    # Delete the temporary directory if it exists
    if os.path.isdir(dest):
        shutil.rmtree(dest)
    # Copy the source directory in the temporary one
    os.makedirs(dest, exist_ok=True)
    filename=os.path.join(dest, "test_suite.tar.gz")
    shutil.copy(src_fullname, filename)
    # and cd into it
    os.chdir(dest)
    if os.path.isfile("test_suite.tar.gz"):
        tar = tarfile.open("test_suite.tar.gz","r:gz")
        tar.extractall()
        tar.close()
        os.remove("test_suite.tar.gz")

    #now we do the different tests one by one

    #Let's do the ALP RGEs
    eftmodels="UnbrokenSM_BFM.fr ALP_EFT.fr"
    uvmodels="UnbrokenSM_BFM.fr ALP_UV.fr"
    mmmodels=uvmodels.split()[-1].replace(".fr","_MM")+" "+eftmodels.split()[-1].replace(".fr","_MM")
    mathematicascript=mmmodels.split()[0]+'.wl'
    try:
        func.create_model(eftmodels.split(), wolframcommand, frpath)
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        func.create_model(uvmodels.split(), wolframcommand, frpath)
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        match_model.match_model(*mmmodels.split())
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        func.compute_rge(*mmmodels.split(), wolframcommand)
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        subprocess.run(wolframcommand+[mathematicascript], check=True)
    except subprocess.CalledProcessError:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')

    #Let's do the matching of the Scalar Singlet on the SMEFT
    uvmodels="UnbrokenSM_BFM.fr Scalar_Singlet_BFM.fr"
    mmmodels="Scalar_Singlet_BFM_MM SMEFT_Green_Bpreserving_MM"
    mathematicascript=mmmodels.split()[0]+'.wl'
    try:
        func.create_model(uvmodels.split(), wolframcommand, frpath)
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        match_model.match_model(*mmmodels.split())
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        func.compute_wilson_coefficients(*mmmodels.split(), wolframcommand)
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        subprocess.run(wolframcommand+[mathematicascript], check=True)
    except subprocess.CalledProcessError:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')

    #Let's do an example of basis change
    uvmodels="UnbrokenSM_BFM.fr newbasis.fr"
    mmmodels="newbasis_MM SMEFT_Green_Bpreserving_MM"
    mathematicascript=mmmodels.split()[0]+'.wl'
    try:
        func.create_model(uvmodels.split(), wolframcommand, frpath)
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        match_model.match_model(*mmmodels.split(), True)
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        func.compute_wilson_coefficients(*mmmodels.split(), wolframcommand)
    except Exception:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')
    try:
        subprocess.run(wolframcommand+[mathematicascript], check=True)
    except subprocess.CalledProcessError:
        ts_up(dest)
        raise SystemExit('Exiting matchmaker-eft. Please check logs in directory .test_suite')


    # put here more tests following the structureabove

    ts_up(dest)
    print ("Tests finished at ",datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"))



import os
import shutil, errno
from distutils.errors import DistutilsFileError
from distutils.dir_util import copy_tree
from pkg_resources import resource_filename
import subprocess
import sys
import requests
import tarfile
import readline
import select
import shlex
from colorama import Fore
from colorama import Style
from . import run_all
from pathlib import Path
from timeit import default_timer as timer
from subprocess import DEVNULL
import itertools
import glob
import re

class Error(Exception):
    """Base class for other exceptions"""
    pass


class BinaryNotFoundError(Error):
    """Raised when we don't find some binary file"""
    pass


def copy(src, dest):
    try:
        shutil.copytree(src, dest)
    except OSError as e:
        # If the error was caused because the source wasn't a directory
        if e.errno == errno.ENOTDIR:
            shutil.copy(src, dest)
        else:
            print('Directory not copied. Error: %s' % e)

def get_path(path):
    return os.path.abspath(resource_filename('matchmakereft', path))

def dress_directory(dirpath):
    return os.path.join(dirpath,"")

def undress_directory(dirpath):
    return os.path.normpath(dirpath)

def list_folder(path):
    """
    Lists folder contents
    """
    if path.startswith(os.path.sep):
        # absolute path
        basedir = os.path.dirname(path)
        contents = os.listdir(basedir)
        # add back the parent
        contents = [os.path.join(basedir, d) for d in contents]
        contents = [d + os.path.sep for d in contents if os.path.isdir(d)]

    else:
        # relative path
        contents = os.listdir(os.curdir)
        basedir = os.path.abspath(os.curdir)
        contents = [os.path.join(basedir, d) for d in contents]
        contents = [d + os.path.sep for d in contents if os.path.isdir(d)]
    return contents

def completer(text, state):
    """
    Our custom completer function
    """
    options = [x for x in list_folder(text) if x.startswith(text)]
    return options[state]


readline.set_completer(completer)

if sys.platform == 'darwin':
    # Apple uses libedit.
    readline.parse_and_bind("bind -e")
    readline.parse_and_bind("bind '\t' rl_complete")
else:
    # Some tweaks for linux
    readline.parse_and_bind('tab: complete')
    readline.set_completer_delims(' \t\n`~!@#$%^&*()=+[{]}\\|;:\'",<>?')


def rep_mysty(file):

    styp=get_path('core')+"/my.sty"
    ln=len(styp)
    mt=ln//2
    fr=styp[:mt]
    ls=styp[mt:ln]
    in_file=open(file, 'r')
    ot_file=open('aux.dat', 'w')
    for line in in_file:
        if 'my.sty' in line:
            ot_file.write(" style= "+ "'" + fr +"'" + "\n")
            ot_file.write("'"+ls +"'"+" ;" + "\n")
        else:
            ot_file.write(line)
    ot_file.close()
    in_file.close()
    shutil.copyfile('aux.dat', file)
    os.remove('aux.dat')


def rp(fname):
    corepre=get_path('core')+"/"
    filename=corepre+fname
    with open(filename) as f:
        s = f.read()
        if 'REPLACEPATH' not in s:
            return

    with open(filename, 'w') as f:
        s = s.replace('REPLACEPATH', dress_directory(get_path('')))
        f.write(s)


def get_feynr_path(file_path, config):
    #print("Checking the Feynrules path")
    # Read the corresponding entry in the config file
    resp = config["USERINFO"]["Feynrules Path"]
    # If this entry is empty, ask for the path and update the config file
    if resp=="" :
        resp = input('Please enter the absolute path to FeynRules: ')
        dir = dress_directory(resp)
        feynrules = os.path.join(dir, "FeynRules.m")
        while not os.path.isfile(feynrules):
            resp=input('FeynRules.m was not found in this path, please enter the absoloute path to FeynRules: ')
            dir = dress_directory(resp)
            feynrules = os.path.join(dir, "FeynRules.m")
        
        config["USERINFO"]["Feynrules Path"] = dir
        with open(file_path, 'w') as conf:
            config.write(conf)
    else:
        # If the entry is not empty, check that it work. If it does not find FR there, ask again and update the config file
        dir = dress_directory(resp)
        feynrules = os.path.join(dir, "FeynRules.m")
        #print("Current Feynrules path is " + dir)
        while not os.path.isfile(feynrules):
            resp=input('FeynRules.m was not found in this path, please enter the absoloute path to FeynRules: ')
            dir = dress_directory(resp)
            feynrules = os.path.join(dir, "FeynRules.m")
            config["USERINFO"]["Feynrules Path"] = dir
            with open(file_path, 'w') as conf:
                config.write(conf)

    # In any case, return a functioning feynrules path
    return(dir)


def rp_feynr(fname, file_path, config):
    corepre=get_path('core')+"/"
    filename=corepre+fname
    with open(filename) as f:
        s = f.read()
        if 'REPLACEPATH' not in s:
            return

    resp = config["USERINFO"]["Feynrules Path"]

    if resp=="" :
        resp = input('Please enter the absolute path to FeynRules: ')
        dir = dress_directory(resp)
        feynrules = os.path.join(dir, "FeynRules.m")
        while not os.path.isfile(feynrules):
            resp=input('FeynRules.m was not found in this path, please enter the absoloute path to FeynRules: ')
            dir = dress_directory(resp)
            feynrules = os.path.join(dir, "FeynRules.m")

        config["USERINFO"]["Feynrules Path"] = dir
        with open(file_path, 'w') as conf:
            config.write(conf)

    dir = dress_directory(resp)
    feynrules = os.path.join(dir, "FeynRules.m")
    while not os.path.isfile(feynrules):
        resp=input('FeynRules.m was not found in this path, please enter the absoloute path to FeynRules: ')
        dir = dress_directory(resp)
        feynrules = os.path.join(dir, "FeynRules.m")
        config["USERINFO"]["Feynrules Path"] = dir
        with open(file_path, 'w') as conf:
            config.write(conf)

    with open(filename, 'w') as f:
        s = s.replace('REPLACEPATH', dir)
        f.write(s)


def rp_model(model_name, filename):
    with open(filename) as f:
        s = f.read()
        if 'MODELPATH' not in s:
            return

    with open(filename, 'w') as f:
        s = s.replace('MODELPATH', dress_directory(model_name))
        f.write(s)


def copy_models(file_path):
    fullname=os.path.join(file_path,"MatchMakerEFT")
    if os.path.isdir(fullname):
        resp=input('Directory '+os.path.join(file_path,"MatchMakerEFT")+' already present, are you sure you want to overwrite the models? [y/N]')
        if (len(resp)==0 or (resp[0] != 'y' and resp[0] != 'Y')):
            return
    src_path=get_path("data")
    src_fullname=os.path.join(src_path,"models.tar.gz")
    os.makedirs(fullname, exist_ok=True)
    filename=os.path.join(fullname, "models.tar.gz")
    shutil.copy(src_fullname, filename)
    if os.path.isfile(filename):
        tar = tarfile.open(filename,"r:gz")
        tar.extractall(fullname)
        tar.close()
        os.remove(filename)
    if os.path.isdir(os.path.join(file_path,"MatchMakerEFT")):
        print("Model directory MatchMakerEFT/ copied to "+file_path)
    else:
        print("It seems like there was a problem copying the files")


def check_executable(code):
    try:
        subprocess.check_output(code)
        #print(code+" properly installed in the system")
    except OSError as e:
        print(code+" does not seem to be in your path, please install it before matching any model")
        raise BinaryNotFoundError


def get_wolfram_command(file_path, config):
    wolframcommand=shlex.split(config["USERINFO"]["Wolfram Script Command"])
    if wolframcommand==[]:
        print("Checking location of Mathematica binaries")
        coredir=get_path("core")
        test_file=os.path.join(coredir,"test.wl")
        commandlist=['wolfram','wolframscript','Wolfram','WolframScrip']
        optionlist=['-file','-script']
        commands=[[a,b] for a in commandlist for b in optionlist]
        for com in commands:
            try:
                subprocess.run(com+[test_file], stdout=DEVNULL, stderr=DEVNULL, timeout=2.5, check=True)
                testout = subprocess.run(com+[test_file], stdout=subprocess.PIPE, stderr=DEVNULL, universal_newlines=True, check=True).stdout
                if testout == '4\n' :
                    wolframcommand=com
                    break
            except (subprocess.TimeoutExpired,FileNotFoundError) as e:
                pass

        # When successful, we add the wolframcommand to the config file
        if wolframcommand!=[]:
            config["USERINFO"]["Wolfram Script Command"]=" ".join(shlex.quote(s) for s in wolframcommand)
            with open(file_path, 'w') as conf:
                config.write(conf)
        # Otherwise, we raise an error
        else:
            print("We can't access the Wolfram Mathematica binaries. Please check that Mathematica is installed and that binaries are included in the user PATH.")
            raise BinaryNotFoundError

    return(wolframcommand)


def fix_math_path(wolframcommand):
    if wolframcommand!=[]:
        coredir=get_path("core")
        getmathpath=os.path.join(coredir,"getmathpath.wl")
        mathoutput=subprocess.run(wolframcommand+[getmathpath], stdout=subprocess.PIPE, universal_newlines=True).stdout
        dir=(mathoutput.split("\n")[0]).replace('"','')
        towrite='$Path = Join[{"'+coredir+'"}, $Path]\n'
        localdir='$Path = Join[{"."}, $Path]\n'
        with open(dir,"r") as f:
            all_lines=f.readlines()
        # we check if the correct path (called towrite) is in init.m
        # if it is remove it (no matter where it is)
        if towrite in all_lines: all_lines.remove(towrite)
        # same for local directory (localdir)
        if localdir in all_lines: all_lines.remove(localdir)
        # now add it to the end of the file
        all_lines.append(towrite)
        all_lines.append(localdir)
        #if coredir not in mathoutput:
        with open(dir.replace('"',''),"w") as myfile:
            for x in all_lines:
                myfile.write(x)#'$Path = Join[{"'+coredir+'"}, $Path]\n')


def create_config_file(file_path, config):
    if not os.path.exists(file_path) or os.stat(file_path).st_size == 0:
        with open(file_path, 'w') as configfile:
            config.write(configfile)


def create_model(model_file, wolframcommand, frpath):
    # model_file is a list ["mod1.fr","mod2.fr",...] with the FeynRules files for the model
    # we will take the name of the last file to name the model
    #os.path.normpath ensures the separators agree with os.sep for any system
    model_file_name=os.path.normpath(model_file[-1][:-3])
    #if model_file_name has the structure "dir/dir1/.../model" model_file_name_local is just "model"
    model_file_name_local=model_file_name.split(os.sep)[-1]
    if wolframcommand!=[]:
        coredir=get_path("core")
        createscript=os.path.join(coredir,"create_MM_model.wl")
        print("Creating model "+model_file_name+"_MM. This might take some time depending on the complexity of the model")
        start = timer()
        try:
            p=subprocess.run(wolframcommand+[createscript]+model_file+[frpath], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
            logfile=model_file_name+".log"
            #logfile=os.path.join(model_file_name+"_MM", model_file_name_local+".log")
            f = open(logfile, "w") 
            f.write(p.stdout.decode())
            f.close()
        except subprocess.CalledProcessError:
            print("\n"+ Fore.RED + "There is a problem with the model, not all particles are defined as light or heavy using FullName->\"light\" or FullName->\"heavy\". Please make sure you define all particles properly and then create the model."+Style.RESET_ALL+"\n")
            raise Exception
        end = timer()
        # check if there is a file called model_file_name.symm
        if os.path.isfile(model_file_name+".symm"):
            shutil.copy(model_file_name+".symm",os.path.join(model_file_name+"_MM","QGRAF","model_data"))
            os.rename(os.path.join(model_file_name+"_MM","QGRAF","model_data",model_file_name_local+".symm"),os.path.join(model_file_name+"_MM","QGRAF","model_data","listareplacesymmetry"))
        # check if there is a file called model_file_name.gauge
        if os.path.isfile(model_file_name+".gauge"):
            shutil.copy(model_file_name+".gauge",os.path.join(model_file_name+"_MM","QGRAF","model_data"))
            os.rename(os.path.join(model_file_name+"_MM","QGRAF","model_data",model_file_name_local+".gauge"),os.path.join(model_file_name+"_MM","QGRAF","model_data","replacegaugedata"))
        # check if there is a file called model_file_name.herm
        if os.path.isfile(model_file_name+".herm"):
            shutil.copy(model_file_name+".herm",os.path.join(model_file_name+"_MM","QGRAF","model_data"))
            os.rename(os.path.join(model_file_name+"_MM","QGRAF","model_data",model_file_name_local+".herm"),os.path.join(model_file_name+"_MM","QGRAF","model_data","listahermiticity"))
        # check if there is a file called model_file_name.red
        if os.path.isfile(model_file_name+".red"):
            shutil.copy(model_file_name+".red",os.path.join(model_file_name+"_MM","QGRAF","model_data"))
            os.rename(os.path.join(model_file_name+"_MM","QGRAF","model_data",model_file_name_local+".red"),os.path.join(model_file_name+"_MM","QGRAF","model_data","redundancies.dat"))
        print("Model "+model_file_name+"_MM created")
        print("It took "+str(int(round(end-start)))+" seconds to create it.")
    else:
        print("can't do it!")


def get_wc(dirname):
    with open(os.path.join(dirname,"newfunctions")) as f:
        s1= f.read().splitlines()
    with open(os.path.join(dirname,"newsymbols")) as f:
        s2= f.read().splitlines()
    return([x for x in s1+s2 if "alpha" in x])

def find_index_ordering(wc,stri):
    if re.search(r'\b'+wc+r'\[\b',stri):
        return ' '.join(re.findall(r'\d',re.search(r'\b'+wc+r'\[(.*?)\]',stri).group(1)))
    else:
        return ' '

def get_vertices(dirnamex):
    dirname=dress_directory(dirnamex)
    # we use reverse sorting so that alphabar comes before alpha and then it is also used.
    wclist=sorted(get_wc(dirname),reverse=True)
    #get list of external particles
    with open(dirname+'/listexternal','r') as the_file:
        externalfields = the_file.read().strip().replace(' ','').split(',')
    midi={i:[] for i in wclist}
    templist=glob.glob(dirname+"v*list")
    for fi in templist:
        with open(fi) as f:
            s=f.read().splitlines()
            for xx in s:
                xxx=xx.split()
                areexternal=[x in externalfields for x in xxx[:-1]]
                for cc in wclist:
                    # if cc is of the form alphaXxx we want to avoid finding
                    # instances of alphaXxxbar so we remove them
                    # also we add them only if all particles are external (as given by the second condition)
                    # the original one with "in" allowed partial match, for instance alphal was found where alphall existed
                    # the new one should get only the correct matches
                    #if cc in xxx[-1].replace(cc+"bar","") and sum(areexternal)==len(areexternal):
                    if re.search(r'\b'+cc+r'\b',xxx[-1].replace(cc+"bar","")) and sum(areexternal)==len(areexternal):
                        midi[cc].append(xxx[:-1]+[find_index_ordering(cc,xxx[-1].replace(cc+"bar",""))])
    return midi


def generate_amplitudes(dirnamex):
    dirname=dress_directory(dirnamex)
    vertices=get_vertices(os.path.join(dirname,"QGRAF","model_data"))
    amplitudelist=[]
    newvertices=[];
    for ke in vertices.keys():
        if len(vertices[ke])>0:
            amplitudelist.append((min(vertices[ke], key=len))[:-1])
            newvertices.append(ke+'  '+(' '.join(min(vertices[ke], key=len)))+'\n')
    amplitudelist.sort()
    finalamplitudelist=sorted(list(amplitudelist for amplitudelist,_ in itertools.groupby(amplitudelist)),key=len)
    f=open(os.path.join(dirname,"QGRAF","model_data","amplitudes.txt"),"w+")
    for am in finalamplitudelist:
        f.write(' '.join(am)+'\n')
    f.close()
    f=open(os.path.join(dirname,"QGRAF","model_data","wc2fields.txt"),"w+")
    for am in newvertices:
        f.write(am)
    f.close()


def generate_allamplitudes(dirnamex):
    dirname=dress_directory(dirnamex)
    vertices=get_vertices(os.path.join(dirname,"QGRAF","model_data"))
    amplitudelist=[]
    newvertices=[];
    for ke in vertices.keys():
        if len(vertices[ke])>0:
            amplitudelist.extend(vertices[ke])
    amplitudelist.sort()
    finalamplitudelist=sorted(list(amplitudelist for amplitudelist,_ in itertools.groupby(amplitudelist)),key=len)
    f=open(os.path.join(dirname,"QGRAF","model_data","allamplitudes.txt"),"w+")
    for am in finalamplitudelist:
        f.write(' '.join(am)+'\n')
    f.close()


def compute_wilson_coefficients(model_file,eft_file, wolframcommand):
    if model_file[-1]=="/":
        mymodel_file=model_file[:-1]
    else:
        mymodel_file=model_file
    if eft_file[-1]=="/":
        myeft_file=eft_file[:-1]
    else:
        myeft_file=eft_file
    if wolframcommand!=[]:
        coredir=get_path("core")
        createscript=os.path.join(coredir,"compute_wilson_coefficients.wl")
        print("Computing Wilson Coefficients for model "+mymodel_file +" matched onto EFT "+myeft_file+". This might take a few minutes depending on the complexity of the model") 
        start = timer()
        subprocess.run(wolframcommand+[createscript, model_file, eft_file])
        end = timer()
        print("Wilson coefficients for Model "+mymodel_file+ " stored in "+mymodel_file+"/MatchingResult.dat")
        print("It took "+str(int(round(end-start)))+" seconds to compute them.")
        # if os.path.isfile(os.path.join(mymodel_file,"WCoutput.txt")):
        #     print("There was a problem with the matching, check details in ",os.path.join(mymodel_file,"WCoutput.txt"))
        # else:
        #     print("The matching was successful")

    else:
        print("can't do it!")



def compute_rge(model_file,eft_file, wolframcommand):
    if model_file[-1]=="/":
        mymodel_file=model_file[:-1]
    else:
        mymodel_file=model_file
    if eft_file[-1]=="/":
        myeft_file=eft_file[:-1]
    else:
        myeft_file=eft_file
    if wolframcommand!=[]:
        coredir=get_path("core")
        createscript=os.path.join(coredir,"compute_rge_model_to_eft.wl")
        print("Computing the RGEs for UV EFT "+mymodel_file +" onto EFT "+myeft_file+". This might take a few minutes depending on the complexity of the model") 
        start = timer()
        subprocess.run(wolframcommand+[createscript, model_file, eft_file])
        end = timer()
        print("It took "+str(int(round(end-start)))+" seconds to compute them.")
        # if os.path.isfile(os.path.join(mymodel_file,"WCoutput.txt")):
        #     print("There was a problem with the matching, check details in ",os.path.join(mymodel_file,"WCoutput.txt"))
        # else:
        #     print("The matching was successful")

    else:
        print("can't do it!")






import re
from . import file as jsf
import subprocess
import glob
import os
import sys
from . import functions as func
from functools import reduce

# This takes a .out file called fi and checks the maximum power of momenta in the amplitude
def get_pnmax_file(fi):
    teo=subprocess.run("grep -E -o 'sSS\^.|sSS' "+fi, shell=True, capture_output=True)
    ou=str(teo.stdout,'utf-8').split('\n')
    return max([0]+[int(x.replace("sSS^","").replace("sSS","1")) for x in ou if len(x)>0])


# This takes a QGRAF model file and extract the list of bosonic particles
def get_bosons(filename):
    with open(filename) as f:
        lines = f.read().splitlines()
    bolist=[]
    for x in lines:
        if ("+]" in x):
            bolist.append(x.split(',')[0].strip().replace("[",""))
            if (x.split(',')[0].strip().replace("[","")!=x.split(',')[1].strip().replace("[","")):
                bolist.append(x.split(',')[1].strip().replace("[",""))
    return bolist


# This takes a MM model called model and checks the maximum dimension of
# the amplitudes computed in that model. It should be used on the EFT model
# after having run run_all on it
def get_maxdim(model):
    modeldir=os.path.join(model,"FORM","proc_0loop")
    bosons=get_bosons(os.path.join(model,'QGRAF','model'))
        

    filelist=glob.glob(os.path.join(modeldir,"*.out"))
    maxdim=0;
    for fi in filelist:
        process=fi.split('/')[-1][:-4]
        particles=process.split('_')
        bparticles=[x for x in particles if x in bosons]
        dim=int((3*len(particles)-len(bparticles))//2)
        nmom=get_pnmax_file(fi)
        maxdim=max(maxdim,dim+nmom)
    print("Maximum dimension for model ",model," is ",maxdim)
    return(maxdim)


# This splits an expression exp that is a sum of monomials into a list of monomials
def split_expression_in_monomials(exp):
    return exp.replace('+','MATCHMAKERMATCHMAKER').replace('-','MATCHMAKERMATCHMAKER').split('MATCHMAKERMATCHMAKER')


# This gets the wilson coefficient and the power of momenta present in monomial exp
def get_wc_momentum_power_monomial(exp):
    alphalist=re.findall(r'\balpha\w*\b',exp)
    if len(alphalist)==1:
        return [alphalist[0],len(re.findall(r'\bpp\d\b',exp))]
    else:
        return['',0]

def get_dimension_wc(model):
    dict={}
    problems=0
    bosons=get_bosons(os.path.join(model,'QGRAF','model'))
    filelist=glob.glob(os.path.join(model,'QGRAF','model_data','v*list'))
    for fi in filelist:
        with open(fi) as f:
            lines = f.read().splitlines()
        for xx in lines:
            particles=xx.split()[:-1]
            bparticles=[x for x in particles if x in bosons]
            dimparts=int((3*len(particles)-len(bparticles))//2)
            fr=split_expression_in_monomials(xx.split()[-1])
            for fri in fr:
                [wc,po]=get_wc_momentum_power_monomial(fri)
                dimwc=4-po-dimparts
                if len(wc)>0:
                    if not wc in dict:
                        dict[wc]=[dimwc]
                    else:
                        if dimwc != dict[wc][-1]:
                            dict[wc].append(dimwc)
                            problems=1
    if problems==0:
        return ['id '+x+' = '+x+'*LAMBDA^('+str(dict[x][0])+');' for x in dict.keys()]
    else:
        return "problem"


def write_wc_dimension(model):
    modeltowrite=[os.path.join(model,'FORM','proc_0loop','wcdimension.dat'),os.path.join(model,'FORM','proc_1loop','wcdimension.dat')]
    for m2w in modeltowrite:
        if not os.path.isfile(m2w):
            fi = open(m2w,'w')
            fi.write('\n'.join(get_dimension_wc(model)))
            fi.close()
    
# This will be used to replace the correct spatial indices
# The indices we are replacing are always inside parenthesis so they are
# either right after [ or right after ,
# by forcing a [ or , before we ensure that no replacement in variable names
# happens
# since we first replace [ with ( we use ( everywhere
def JSP_replace(st,ilist,plist,indexlist,dummyrepl):
    repl=[('[','('),(']',')')]
    for x in dummyrepl:
        repl.extend([x])
    for x in indexlist.replace(" ","").replace("\n","").split(","):
        repl.extend([('('+x+x+str(i+1),'('+x+str(t)) for (i,t) in enumerate(ilist)])
        repl.extend([(','+x+x+str(i+1),','+x+str(t)) for (i,t) in enumerate(ilist)])
    repl.extend([('(pp'+str(i+1),'('+t) for (i,t) in enumerate(plist)])
    repl.extend([(',pp'+str(i+1),','+t) for (i,t) in enumerate(plist)])
    return  reduce(lambda a, kv: a.replace(*kv), repl, st)


def allpos(i):
    res=int(i)
    if res<0:
        res += 100
    return str(res)


def feynman_rules(dirname):
    # This is a dictionary of dictionaries encoding all the Feynman rules
    alldicts={}
    templist=glob.glob(dirname+"QGRAF/model_data/*list")
    allkeys=[x.split('/')[-1][:-4] for x in templist]
    for key in allkeys:
        filename=dirname+"QGRAF/model_data/"+key+"list"
        tempdict={}
        for line in open(filename):
            tempdict[" ".join(line.split()[:-1])]=line.split()[-1]
        alldicts[key]=tempdict
    return alldicts


def read_new_params(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/listparam','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace('\n','') for x in all_lines])


def read_new_functions(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newfunctions','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace('\n','') for x in all_lines])


def read_new_gauge_functions(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newgaugefunctions','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace('\n','') for x in all_lines])


def read_new_flavor_functions(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newflavorfunctions','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace('\n','') for x in all_lines])


def read_new_flavor_mass_functions(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newflavormassfunctions','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace('\n','') for x in all_lines])


def read_new_symbols(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newsymbols','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace('\n','') for x in all_lines])


def read_new_indices(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newindices','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace("\n","").split("=")[0] for x in all_lines])


def read_new_heavy_flavor_indices(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newheavyflavorindices','r') as f:
        all_lines = f.readlines()
    return [x.replace("\n","").split("=")[0] for x in all_lines]


def read_new_indices_with_dims(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newindices','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace('\n','') for x in all_lines])


def read_new_gauge_indices_with_dims(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newgaugeindices','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.replace('\n','').replace('\t','=') for x in all_lines])


def read_light_masses(dirname):
    # define light masses
    with open(dirname+'QGRAF/model_data/listlightmass','r') as f:
        all_lines = f.readlines()
    answer=""
    if len(all_lines)>0:
        answer=', '.join(all_lines[0].split())
    return answer


def read_heavy_masses(dirname):
    # define heavy masses
    with open(dirname+'QGRAF/model_data/listheavymass','r') as f:
        all_lines = f.readlines()
    answer=""
    if len(all_lines)>0:
        answer=', '.join(all_lines[0].split())
    return answer


def write_aux_files(dirname,filename,alldicts,maxdim):
    process=filename.split('/')[-1][:-4]
    particles=process.split('_')
    bosons=get_bosons(dirname+'QGRAF/model')
    bparticles=[x for x in particles if x in bosons]
    dim=int((3*len(particles)-len(bparticles))//2)
    thenmax=maxdim-dim
    looporder=filename.split('/')[-2]
    formdir=dirname+'FORM/'+looporder+'/'
    if not os.path.isdir(formdir):
        subprocess.call("mkdir "+formdir,shell=True)

    newindices=read_new_indices(dirname)
    listnewindices=[x for x in newindices.replace(" ","").split(',')]
    diagnum=1
    lastline=0
    outputline=''

    outputfilename=formdir+process+'.aux' # one diagram at a time

    if not os.path.isfile(outputfilename):
        of = open(outputfilename,"w")
        dummyi=0
        dummymumu=0
        dummies={x:0 for x in listnewindices}

        for line in open(filename):
            dummyi+=1
            datalinetemp=line.rstrip().replace('(',',').replace(')','').replace('*','').split(',')
            if datalinetemp[-1]=='' and len(datalinetemp)>1:
                # last line of the current diagram
                lastline=1
                dataline=datalinetemp[:-1]
            else:
                dataline=datalinetemp

            # Now we start with the proper replacements
            if dataline[0]=='' and len(dataline)>1:
                # numerical factor
                outputline += "("+dataline[1].replace("+","")+")"
            elif len(dataline)>1:
                [op,fields,indices0,ppi]=dataline[0].strip(),dataline[1::3],dataline[2::3],dataline[3::3]
                indices = [allpos(ii) for ii in indices0]
                dummyrepl=[]
                for xx in listnewindices:
                    if (xx+xx+"minus" in alldicts[op][" ".join(fields)]):
                        maxhere=max([x for x in re.findall(r"[\w']+",alldicts[op][" ".join(fields)].replace("mumumu","mumu")) if xx+xx+"minus" in x]).replace(xx+xx+"minus","")
                        for i in range(int(maxhere)):
                            dummyrepl.append((xx+xx+"minus"+str(i+1),xx+"minus"+str(i+1+dummies[xx])))
                        dummies[xx]=dummies[xx]+int(maxhere)

                outputline += "*("+JSP_replace(alldicts[op][" ".join(fields)],indices,ppi,newindices,dummyrepl)+")"

            if lastline==1:
                of.write("id diags"+str(diagnum)+" = "+outputline+";\n\n")
                diagnum += 1
                outputline=''
                lastline = 0
                dummyi=0
                dummymumu=0
                dummies={x:0 for x in listnewindices}


        of.close()
    return diagnum,process,thenmax,formdir


def generate_index_definition(newindices,newgaugeindices,form_directory):

    data=""
    filelist=glob.glob(os.path.join(form_directory,"*aux"))
    for filename in filelist:
        with open(filename) as fi:
            data+=fi.read()
    theindices=[x.split("=") for x in newindices.replace(" ","").split(',')]

    thegaugeindices=[]
    if len(newgaugeindices) >0:
        thegaugeindices=[x.split("=") for x in newgaugeindices.replace(" ","").split(',')] 
    
    stri="i <THEINDEX1=INDEXDIM>, ... , <THEINDEX100=INDEXDIM>, <THEINDEXminus1=INDEXDIM>, ... , <THEINDEXminus100=INDEXDIM>THEEXTRAINDICES;\nset THEINDEXi: THEINDEX1, ... , THEINDEX100, THEINDEXminus1, ... , THEINDEXminus100THE2EXTRAINDICES;\n\n"
    defs=""
    for x in theindices:
        te=re.findall(r'[^a-zA-Z]'+x[0]+x[0]+r'[^(inus)][a-zA-Z]*\d+',data)
        if len(te)>0:
            finalte=list(set([re.sub(r'\W+','',xx) for xx in te]))
            finaltee=[xx+'='+x[1] for xx in finalte]
            finaltee2=', '.join(finalte)
            if len(finaltee)>0:
                requete=', '.join(finaltee)
                THEEXTRAINDICESSUS=', '+requete
                THE2EXTRAINDICESSUS=', '+finaltee2
            else:
                THEEXTRAINDICESSUS=''
                THE2EXTRAINDICESSUS=''
        else:
            THEEXTRAINDICESSUS=''
            THE2EXTRAINDICESSUS=''
        defs +=stri.replace("THEINDEX",x[0]).replace("INDEXDIM",x[1]).replace("THEEXTRAINDICES",THEEXTRAINDICESSUS).replace("THE2EXTRAINDICES",THE2EXTRAINDICESSUS)
    for x in [x for x  in thegaugeindices if x not in theindices]:
        te=re.findall(r'[^a-zA-Z]'+x[0]+x[0]+r'[^(inus)][a-zA-Z]*\d+',data)
        if len(te)>0:
            finalte=list(set([re.sub(r'\W+','',xx) for xx in te]))
            finaltee=[xx+'='+x[1] for xx in finalte]
            finaltee2=', '.join(finalte)
            if len(finaltee)>0:
                requete=', '.join(finaltee)
                THEEXTRAINDICESSUS=', '+requete
                THE2EXTRAINDICESSUS=', '+finaltee2
            else:
                THEEXTRAINDICESSUS=''
                THE2EXTRAINDICESSUS=''
        else:
            THEEXTRAINDICESSUS=''
            THE2EXTRAINDICESSUS=''
        defs +=stri.replace("THEINDEX",x[0]).replace("INDEXDIM",x[1]).replace("THEEXTRAINDICES",THEEXTRAINDICESSUS).replace("THE2EXTRAINDICES",THE2EXTRAINDICESSUS)


    return(defs)


def generate_heavy_flavor_deltas(newheavyflavorindices):
    strif="id deltaF(mu1?THEINDEXi,mu2?THEINDEXi)=deltaFF(mu1,mu2);\n"
    defs=""
    for x in newheavyflavorindices:
        defs +=strif.replace("THEINDEX",x)
    return(defs)


def generate_id_deltas(newindices,newgaugeindices):
    theindices=[x.split("=") for x in newindices.replace(" ","").split(',')]
    thegaugeindices=[]
    if len(newgaugeindices) >0:
        thegaugeindices=[x.split("=") for x in newgaugeindices.replace(" ","").split(',')] 
    theflavorindices=[x for x in theindices if x not in thegaugeindices]
    stri="id d_(mu1?THEINDEXi,mu2?THEINDEXi)=dd(INDEXDIM,mu1,mu2);\n"
    strif="id d_(mu1?THEINDEXi,mu2?THEINDEXi)=ddF(INDEXDIM,mu1,mu2);\n"
    defs=""
    for x in thegaugeindices:
        defs +=stri.replace("THEINDEX",x[0]).replace("INDEXDIM",x[1])
    for x in theflavorindices:
        defs +=strif.replace("THEINDEX",x[0]).replace("INDEXDIM",x[1])
    return(defs)


def generate_call_dummy_indices(newindices):
    theindices=[x.split("=") for x in newindices.replace(" ","").split(',')]
    stri="#call dummyindices(THEINDEXi,THEINDEX)\n"
    defs=""
    for x in theindices:
        defs +=stri.replace("THEINDEX",x[0])
    return(defs)


def generate_form(dirname0,isEFT,maxdim):
    corepre=func.get_path('core')+"/"
    if os.path.isdir(dirname0):
        dirname=jsf.proper_dir_path(dirname0)
        write_wc_dimension(dirname)


        alldicts=feynman_rules(dirname)
        # Go through all the QGRAF files
        if isEFT:
            # EFT do only tree level
            filelist=glob.glob(dirname+"QGRAF/proc_0loop/*qgf")
        else:
            filelist=glob.glob(dirname+"QGRAF/proc_*loop/*qgf")
        for filename in filelist:
            diagnum,process,thenmax,formdir=write_aux_files(dirname,filename,alldicts,maxdim)
            numparticles=len(process.split("_"))
            # check if form file is there, if it is do not do anything
            if not os.path.isfile(formdir+process+'.frm'):
                ## Let's now write the form skeleton program
                myrepl= ("THELAMBDAINVERSEMAX",str(3-int(maxdim))),('COREPRE',corepre),('MM_NUMBER_OF_PARTICLES',str(numparticles)),('MM_NUMBER_OF_DIAGRAMS',str(diagnum-1)),("THENMAX",str(thenmax)),("FILE_WITH_DIAGRAMS",process),("PROCESS",process.replace("_","").replace("[","").replace("]",""))
                jsf.dress_skeleton(myrepl, corepre+'skeleton_parallel_new.frm',formdir+process+'.frm')

        newgaugefunctions=read_new_gauge_functions(dirname)
        newflavorfunctions=read_new_flavor_functions(dirname)
        newflavormassfunctions=read_new_flavor_mass_functions(dirname)
        newfunctions=read_new_functions(dirname)
        newmassfunctions=list(set(newfunctions.replace(" ","").split(","))-set(newflavorfunctions.replace(" ","").split(","))-set(newgaugefunctions.replace(" ","").split(",")))
        newsymbols=read_new_symbols(dirname)
        if len(newmassfunctions)>0:
            newsymbols=newsymbols+";\nCF "+", ".join(newmassfunctions)
        newindices=read_new_indices_with_dims(dirname)
        newheavyflavorindices=read_new_heavy_flavor_indices(dirname)
        newgaugeindices=read_new_gauge_indices_with_dims(dirname)
        heavy_flavor_deltas=generate_heavy_flavor_deltas(newheavyflavorindices)
        id_deltas=generate_id_deltas(newindices,newgaugeindices)
        call_dummy_indices=generate_call_dummy_indices(newindices)
        lightmasses=read_light_masses(dirname)
        heavymasses=read_heavy_masses(dirname)
        # isnoteft is 0  (false) if there are no heavy masses (and therefore is an eft)
        # and different from 0 (true) if there are heavy masses (and therefore is not an eft)
        # note that on RGEMaker mode for the 1 loop calculation
        # isEFT will be False but isnoteft should be also false (0)
        # If isnoteft='0' then 4D properties of the gamma matrices are used
        # this should be used only in RGEMaker mode
        # we set it to a non-zero value and then, if we have to do one-loop calculations
        # we fix it to its real value (0 for RGEMaker mode and something different from 0 for MatchMaker mode)
        isnoteft='1'
        # all the aux files have been written already, we can get the extra indices used now
        if isEFT:
            # EFT, do only tree level
            formdirlist=glob.glob(dirname+"FORM/proc_0*")
        else:
            formdirlist=glob.glob(dirname+"FORM/proc_*")
            isnoteft=str(len(heavymasses))
        if (len(formdirlist) == 2):
            extrasymbol="esfull"
        else:
            extrasymbol="eseft"
        for fdir in formdirlist:
            loopord=fdir.split('_')[-1].replace('loop','')
            indexdefinition=generate_index_definition(newindices,newgaugeindices,fdir)
            myrepl= ('DUMMY_INDICES',call_dummy_indices),('ID_DELTAS',id_deltas),('HF_DELTAS',heavy_flavor_deltas),('NEW_SYMBOLS',newsymbols),('NEW_GAUGE_FUNCTIONS',newgaugefunctions),('NEW_FLAVOR_FUNCTIONS',newflavorfunctions),('NEW_FLAVOR_MASS_FUNCTIONS',newflavormassfunctions),("INDEX_DEFINITION",indexdefinition),('EXTRA_SYMBOLS',extrasymbol),('LIGHTMASSESSET',lightmasses),('HEAVYMASSESSET',heavymasses),('ISNOTEFT',isnoteft),('LOOPORDER',loopord)
            #print(fdir+'/generaldefs.h')
            jsf.dress_skeleton(myrepl, corepre+'generaldefs_skeleton.h',fdir+'/generaldefs.h')
    else:
        print("not a valid directory")



import subprocess
import glob
import sys
import os
import re
import math
from colorama import Fore
from colorama import Style
from tqdm.auto import tqdm, trange
from time import sleep
from . import file as jsf

def better_log(file,indexlist):
    if not os.path.isfile(file+".mat"):
        with open(file) as f:
            lines = f.read().splitlines()
            i0=0
            i1=0
            for i,li in enumerate(lines):
                if "ampl" in li:
                    i0=i;
                if ";" in li:
                    i1=i;
        lin=''.join([x.strip() for x in  lines[i0:i1+1]])
        for i in range(20):
            lin=lin.replace("f"+str(i+1),"fl"+str(i+1))
        lin=lin.replace("e_","ee")
        lin=lin.replace("[MM.epsi]","epsilonbar")
        lin=re.sub(r'd_\(([0-9a-zA-Z]*),([0-9a-zA-Z]*)\)', r'dd[\1,\2]', lin)
        lin=re.sub(r'esfull\(([0-9]*)\)', r'esfull[\1]', lin)
        lin=re.sub(r'eseft\(([0-9]*)\)', r'eseft[\1]', lin)
        lin=lin.replace("=","->")
        lin=lin.replace(";",",")
        lin="{"+lin+"}"
        lin=lin.replace(",}","}")
        with open(file+".mat","w") as ff:
            ff.write(lin)

def read_new_indices(dirname):
    # define new parameters
    with open(dirname+'QGRAF/model_data/newindices','r') as f:
        all_lines = f.readlines()
    return ', '.join([x.split("=")[0] for x in all_lines])

def run_form(dirname0,isEFT):
    if os.path.isdir(dirname0):
        filesdonealready=[]
        dirname=jsf.proper_dir_path(dirname0)
        newindices=read_new_indices(dirname)
        currentdir=os.getcwd()
        if os.path.isdir(dirname+"FORM/proc_0loop/"):
            os.chdir(dirname+"FORM/proc_0loop/")
            aux = glob.glob("*frm")
            nam = str(math.ceil(math.log10(len(aux))))
            frm = "{n_fmt:"+nam+"}/{total_fmt:"+nam+"} amplitudes | {percentage:3.0f}% |{bar:20}| {desc}" 
            filelist=tqdm(aux, bar_format=frm)
            filelist.write("Computing the tree-level amplitudes for model "+dirname[:-1])
            for fil in filelist:

                filelist.set_description_str("(amplitude %s)" % fil[:-4].replace("_"," "))
                # filelist.set_description("Computing the tree-level amplitude %s" % fil[:-4].replace("_"," "))
                # we check for the .out file because the .log is created even if form doesnt run smoothly
                if os.path.isfile(fil[:-4]+".out"):
                    filesdonealready.append(fil[:-4]+".out")
                else:

                    try:
                        p=subprocess.run(("form -F "+fil).split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
                    except subprocess.CalledProcessError as E:
                        # there was a problem, go back to the original directory and raise an exception
                        os.chdir(currentdir)
                        raise Exception("\n"+ Fore.RED + "There was an error when running "+' '.join(E.args[1])+" in "+dirname+"FORM/proc_0loop/, please check!"+Style.RESET_ALL+"\n")
                if os.path.isfile(fil[:-4]+".log"):
                    better_log(fil[:-4]+".log",newindices)
                if os.path.isfile(fil[:-4]+".out"):
                    better_log(fil[:-4]+".out",newindices)
            os.chdir(currentdir)
        if os.path.isdir(dirname+"FORM/proc_1loop/") and not isEFT:
            os.chdir(dirname+"FORM/proc_1loop/")
            aux = glob.glob("*frm")
            nam = str(math.ceil(math.log10(len(aux))))
            frm = "{n_fmt:"+nam+"}/{total_fmt:"+nam+"} amplitudes | {percentage:3.0f}% |{bar:20}| {desc}" 
            filelist=tqdm(aux, bar_format=frm)
            filelist.write("Computing the one-loop amplitudes for model "+dirname[:-1])
            for fil in filelist:
                filelist.set_description_str("(amplitude %s)" % fil[:-4].replace("_"," "))
                # we check for the .out file because the .log is created even if form doesnt run smoothly
                if os.path.isfile(fil[:-4]+".out"):
                    filesdonealready.append(fil[:-4]+".out")
                else:
                    try:
                        p=subprocess.run(("form -F "+fil).split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
                    except subprocess.CalledProcessError as E:
                        # there was a problem, go back to the original directory and raise an exception
                        os.chdir(currentdir)
                        raise Exception("\n"+ Fore.RED + "There was an error when running "+' '.join(E.args[1])+" in "+dirname+"FORM/proc_1loop/, please check!"+Style.RESET_ALL+"\n")
                if os.path.isfile(fil[:-4]+".log"):
                    better_log(fil[:-4]+".log",newindices)
                if os.path.isfile(fil[:-4]+".out"):
                    better_log(fil[:-4]+".out",newindices)
            os.chdir(currentdir)

    else:
        print("not a valid directory")

